@@ -70,8 +70,10 @@ class JSONPathEnvironment:
     ## Class attributes
 
     Arguments:
-        filter_caching (bool): If `True`, filter expressions will be cached where
-            possible.
+        filter_caching (bool): If `True`, filter expressions will be cached
+            where possible.
+        unicode_escape: If `True`, decode UTF-16 escape sequences found in
+            JSONPath string literals.
 
     Attributes:
         filter_context_token (str): The pattern used to select extra filter context
@@ -113,8 +115,18 @@ class JSONPathEnvironment:
     parser_class: Type[Parser] = Parser
     match_class: Type[JSONPathMatch] = JSONPathMatch
 
-    def __init__(self, *, filter_caching: bool = True) -> None:
-        self.filter_caching = filter_caching
+    def __init__(
+        self,
+        *,
+        filter_caching: bool = True,
+        unicode_escape: bool = True,
+    ) -> None:
+        self.filter_caching: bool = filter_caching
+        """Enable or disable filter expression caching."""
+
+        self.unicode_escape: bool = unicode_escape
+        """Enable or disable decoding of UTF-16 escape sequences found in
+        JSONPath string literals."""
 
         self.lexer: Lexer = self.lexer_class(env=self)
         """The lexer bound to this environment."""
@@ -137,6 +149,11 @@ class JSONPathEnvironment:
             A `JSONPath` or `CompoundJSONPath`, ready to match against some data.
                 Expect a `CompoundJSONPath` if the path string uses the _union_ or
                 _intersection_ operators.
+
+        Raises:
+            JSONPathSyntaxError: If _path_ is invalid.
+            JSONPathTypeError: If filter functions are given arguments of an
+                unacceptable type.
         """
         tokens = self.lexer.tokenize(path)
         stream = TokenStream(tokens)
@@ -156,7 +173,7 @@ class JSONPathEnvironment:
 
                 if stream.current.kind == TOKEN_UNION:
                     stream.next_token()
-                    _path.union(
+                    _path = _path.union(
                         JSONPath(
                             env=self,
                             selectors=self.parser.parse(stream),
@@ -164,7 +181,7 @@ class JSONPathEnvironment:
                     )
                 elif stream.current.kind == TOKEN_INTERSECTION:
                     stream.next_token()
-                    _path.intersection(
+                    _path = _path.intersection(
                         JSONPath(
                             env=self,
                             selectors=self.parser.parse(stream),
